# -*- python-mode -*-

## Copyright (C) 2020 Peter Patel-Schneider
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import ctypes as _ctypes
import logging
import os as _os
import os.path as _path
import platform as _platform
import sys as _sys
import time as _time

from math import sqrt as _sqrt
from struct import unpack as _unpack

# There is no evdev on macOS or Windows. Diversion will not work without
# it but other Solaar functionality is available.
if _platform.system() in ('Darwin', 'Windows'):
    evdev = None
else:
    import evdev

import dbus
import keysyms.keysymdef as _keysymdef
import psutil

from yaml import add_representer as _yaml_add_representer
from yaml import dump_all as _yaml_dump_all
from yaml import safe_load_all as _yaml_safe_load_all

from .common import NamedInt
from .device import Device as _Device
from .hidpp20 import FEATURE as _F
from .special_keys import CONTROL as _CONTROL

import gi  # isort:skip

gi.require_version('Gdk', '3.0')  # isort:skip
from gi.repository import Gdk, GLib  # NOQA: E402 # isort:skip

logger = logging.getLogger(__name__)

#
# See docs/rules.md for documentation
#
# Several capabilities of rules depend on aspects of GDK, X11, or XKB
# As the Solaar GUI uses GTK, Glib and GDK are always available and are obtained from gi.repository
#
# Process condition depends on X11 from python-xlib, and is probably not possible at all in Wayland
# MouseProcess condition depends on X11 from python-xlib, and is probably not possible at all in Wayland
# Modifiers condition depends only on GDK
# KeyPress action determines whether a keysym is a currently-down modifier using get_modifier_mapping from python-xlib;
#   under Wayland no modifier keys are considered down so all modifier keys are pressed, potentially leading to problems
# KeyPress action translates key names to keysysms using the local file described for GUI keyname determination
# KeyPress action gets the current keyboard group using XkbGetState from libX11.so using ctypes definitions
#   under Wayland the keyboard group is None resulting in using the first keyboard group
# KeyPress action translates keysyms to keycodes using the GDK keymap
# KeyPress, MouseScroll, and MouseClick actions use XTest (under X11) or uinput.
# For uinput to work the user must have write access for /dev/uinput.
# To get this access run  sudo setfacl -m u:${user}:rw /dev/uinput
#
# Rule GUI keyname determination uses a local file generated
#   from http://cgit.freedesktop.org/xorg/proto/x11proto/plain/keysymdef.h
#   and http://cgit.freedesktop.org/xorg/proto/x11proto/plain/XF86keysym.h
# because there does not seem to be a non-X11 file for this set of key names

# Setting up is complex because there are several systems that each provide partial facilities:
# GDK - always available (when running with a window system) but only provides access to keymap
# X11 - provides access to active process and process with window under mouse and current modifier keys
# Xtest extension to X11 - provides input simulation, partly works under Wayland
# Wayland - provides input simulation

XK_KEYS = _keysymdef.keysymdef

# Event codes - can't use Xlib.X codes because Xlib might not be available
_KEY_RELEASE = 0
_KEY_PRESS = 1
_BUTTON_RELEASE = 2
_BUTTON_PRESS = 3

CLICK, DEPRESS, RELEASE = 'click', 'depress', 'release'

gdisplay = Gdk.Display.get_default()  # can be None if Solaar is run without a full window system
gkeymap = Gdk.Keymap.get_for_display(gdisplay) if gdisplay else None
if logger.isEnabledFor(logging.INFO):
    logger.info('GDK Keymap %sset up', '' if gkeymap else 'not ')

wayland = _os.getenv('WAYLAND_DISPLAY')  # is this Wayland?
if wayland:
    logger.warning(
        'rules cannot access modifier keys in Wayland, '
        'accessing process only works on GNOME with Solaar Gnome extension installed'
    )

try:
    import Xlib
    _x11 = None  # X11 might be available
except Exception:
    _x11 = False  # X11 is not available

xtest_available = True  # Xtest might be available
xdisplay = None
Xkbdisplay = None  # xkb might be available
modifier_keycodes = []
XkbUseCoreKbd = 0x100

_dbus_interface = None


class XkbDisplay(_ctypes.Structure):
    """ opaque struct """


class XkbStateRec(_ctypes.Structure):
    _fields_ = [('group', _ctypes.c_ubyte), ('locked_group', _ctypes.c_ubyte), ('base_group', _ctypes.c_ushort),
                ('latched_group', _ctypes.c_ushort), ('mods', _ctypes.c_ubyte), ('base_mods', _ctypes.c_ubyte),
                ('latched_mods', _ctypes.c_ubyte), ('locked_mods', _ctypes.c_ubyte), ('compat_state', _ctypes.c_ubyte),
                ('grab_mods', _ctypes.c_ubyte), ('compat_grab_mods', _ctypes.c_ubyte), ('lookup_mods', _ctypes.c_ubyte),
                ('compat_lookup_mods', _ctypes.c_ubyte),
                ('ptr_buttons', _ctypes.c_ushort)]  # something strange is happening here but it is not being used


def x11_setup():
    global _x11, xdisplay, modifier_keycodes, NET_ACTIVE_WINDOW, NET_WM_PID, WM_CLASS, xtest_available
    if _x11 is not None:
        return _x11
    try:
        from Xlib.display import Display
        xdisplay = Display()
        modifier_keycodes = xdisplay.get_modifier_mapping()  # there should be a way to do this in Gdk
        NET_ACTIVE_WINDOW = xdisplay.intern_atom('_NET_ACTIVE_WINDOW')
        NET_WM_PID = xdisplay.intern_atom('_NET_WM_PID')
        WM_CLASS = xdisplay.intern_atom('WM_CLASS')
        _x11 = True  # X11 available
        if logger.isEnabledFor(logging.INFO):
            logger.info('X11 library loaded and display set up')
    except Exception:
        logger.warning('X11 not available - some rule capabilities inoperable', exc_info=_sys.exc_info())
        _x11 = False
        xtest_available = False
    return _x11


def gnome_dbus_interface_setup():
    global _dbus_interface
    if _dbus_interface is not None:
        return _dbus_interface
    try:
        bus = dbus.SessionBus()
        remote_object = bus.get_object('org.gnome.Shell', '/io/github/pwr_solaar/solaar')
        _dbus_interface = dbus.Interface(remote_object, 'io.github.pwr_solaar.solaar')
    except dbus.exceptions.DBusException:
        logger.warning('Solaar Gnome extension not installed - some rule capabilities inoperable', exc_info=_sys.exc_info())
        _dbus_interface = False
    return _dbus_interface


def xkb_setup():
    global X11Lib, Xkbdisplay
    if Xkbdisplay is not None:
        return Xkbdisplay
    try:  # set up to get keyboard state using ctypes interface to libx11
        X11Lib = _ctypes.cdll.LoadLibrary('libX11.so')
        X11Lib.XOpenDisplay.restype = _ctypes.POINTER(XkbDisplay)
        X11Lib.XkbGetState.argtypes = [_ctypes.POINTER(XkbDisplay), _ctypes.c_uint, _ctypes.POINTER(XkbStateRec)]
        Xkbdisplay = X11Lib.XOpenDisplay(None)
        if logger.isEnabledFor(logging.INFO):
            logger.info('XKB display set up')
    except Exception:
        logger.warning('XKB display not available - rules cannot access keyboard group', exc_info=_sys.exc_info())
        Xkbdisplay = False
    return Xkbdisplay


if evdev:
    buttons = {
        'unknown': (None, None),
        'left': (1, evdev.ecodes.ecodes['BTN_LEFT']),
        'middle': (2, evdev.ecodes.ecodes['BTN_MIDDLE']),
        'right': (3, evdev.ecodes.ecodes['BTN_RIGHT']),
        'scroll_up': (4, evdev.ecodes.ecodes['BTN_4']),
        'scroll_down': (5, evdev.ecodes.ecodes['BTN_5']),
        'scroll_left': (6, evdev.ecodes.ecodes['BTN_6']),
        'scroll_right': (7, evdev.ecodes.ecodes['BTN_7']),
        'button8': (8, evdev.ecodes.ecodes['BTN_8']),
        'button9': (9, evdev.ecodes.ecodes['BTN_9']),
    }

    # uinput capability for keyboard keys, mouse buttons, and scrolling
    key_events = [c for n, c in evdev.ecodes.ecodes.items() if n.startswith('KEY') and n != 'KEY_CNT']
    for (_, evcode) in buttons.values():
        if evcode:
            key_events.append(evcode)
    devicecap = {evdev.ecodes.EV_KEY: key_events, evdev.ecodes.EV_REL: [evdev.ecodes.REL_WHEEL, evdev.ecodes.REL_HWHEEL]}
else:
    # Just mock these since they won't be useful without evdev anyway
    buttons = {}
    key_events = []
    devicecap = {}

udevice = None


def setup_uinput():
    global udevice
    if udevice is not None:
        return udevice
    try:
        udevice = evdev.uinput.UInput(events=devicecap, name='solaar-keyboard')
        if logger.isEnabledFor(logging.INFO):
            logger.info('uinput device set up')
        return True
    except Exception as e:
        logger.warning('cannot create uinput device: %s', e)


if wayland:  # Wayland can't use xtest so may as well set up uinput now
    setup_uinput()


def kbdgroup():
    if xkb_setup():
        state = XkbStateRec()
        X11Lib.XkbGetState(Xkbdisplay, XkbUseCoreKbd, _ctypes.pointer(state))
        return state.group
    else:
        return None


def modifier_code(keycode):
    if wayland or not x11_setup() or keycode == 0:
        return None
    for m in range(0, len(modifier_keycodes)):
        if keycode in modifier_keycodes[m]:
            return m


key_down = None
key_up = None


def signed(bytes):
    return int.from_bytes(bytes, 'big', signed=True)


def xy_direction(_x, _y):
    # normalize x and y
    m = _sqrt((_x * _x) + (_y * _y))
    if m == 0:
        return 'noop'
    x = round(_x / m)
    y = round(_y / m)
    if x < 0 and y < 0:
        return 'Mouse Up-left'
    elif x > 0 and y < 0:
        return 'Mouse Up-right'
    elif x < 0 and y > 0:
        return 'Mouse Down-left'
    elif x > 0 and y > 0:
        return 'Mouse Down-right'
    elif x > 0:
        return 'Mouse Right'
    elif x < 0:
        return 'Mouse Left'
    elif y > 0:
        return 'Mouse Down'
    elif y < 0:
        return 'Mouse Up'
    else:
        return 'noop'


def simulate_xtest(code, event):
    global xtest_available
    if x11_setup() and xtest_available:
        try:
            event = (
                Xlib.X.KeyPress if event == _KEY_PRESS else Xlib.X.KeyRelease if event == _KEY_RELEASE else
                Xlib.X.ButtonPress if event == _BUTTON_PRESS else Xlib.X.ButtonRelease if event == _BUTTON_RELEASE else None
            )
            Xlib.ext.xtest.fake_input(xdisplay, event, code)
            xdisplay.sync()
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('xtest simulated input %s %s %s', xdisplay, event, code)
            return True
        except Exception as e:
            xtest_available = False
            logger.warning('xtest fake input failed: %s', e)


def simulate_uinput(what, code, arg):
    global udevice
    if setup_uinput():
        try:
            udevice.write(what, code, arg)
            udevice.syn()
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('uinput simulated input %s %s %s', what, code, arg)
            return True
        except Exception as e:
            udevice = None
            logger.warning('uinput write failed: %s', e)


def simulate_key(code, event):  # X11 keycode but Solaar event code
    if not wayland and simulate_xtest(code, event):
        return True
    if simulate_uinput(evdev.ecodes.EV_KEY, code - 8, event):
        return True
    logger.warning('no way to simulate key input')


def click_xtest(button, count):
    if isinstance(count, int):
        for _ in range(count):
            if not simulate_xtest(button[0], _BUTTON_PRESS):
                return False
            if not simulate_xtest(button[0], _BUTTON_RELEASE):
                return False
    else:
        if count != RELEASE:
            if not simulate_xtest(button[0], _BUTTON_PRESS):
                return False
        if count != DEPRESS:
            if not simulate_xtest(button[0], _BUTTON_RELEASE):
                return False
    return True


def click_uinput(button, count):
    if isinstance(count, int):
        for _ in range(count):
            if not simulate_uinput(evdev.ecodes.EV_KEY, button[1], 1):
                return False
            if not simulate_uinput(evdev.ecodes.EV_KEY, button[1], 0):
                return False
    else:
        if count != RELEASE:
            if not simulate_uinput(evdev.ecodes.EV_KEY, button[1], 1):
                return False
        if count != DEPRESS:
            if not simulate_uinput(evdev.ecodes.EV_KEY, button[1], 0):
                return False
    return True


def click(button, count):
    if not wayland and click_xtest(button, count):
        return True
    if click_uinput(button, count):
        return True
    logger.warning('no way to simulate mouse click')
    return False


def simulate_scroll(dx, dy):
    if not wayland and xtest_available:
        success = True
        if dx:
            success = click_xtest(buttons['scroll_right' if dx > 0 else 'scroll_left'], count=abs(dx))
        if dy and success:
            success = click_xtest(buttons['scroll_up' if dy > 0 else 'scroll_down'], count=abs(dy))
        if success:
            return True
    if setup_uinput():
        success = True
        if dx:
            success = simulate_uinput(evdev.ecodes.EV_REL, evdev.ecodes.REL_HWHEEL, dx)
        if dy and success:
            success = simulate_uinput(evdev.ecodes.EV_REL, evdev.ecodes.REL_WHEEL, dy)
        if success:
            return True
    logger.warning('no way to simulate scrolling')


def thumb_wheel_up(f, r, d, a):
    global thumb_wheel_displacement
    if f != _F.THUMB_WHEEL or r != 0:
        return False
    if a is None:
        return signed(d[0:2]) < 0 and signed(d[0:2])
    elif thumb_wheel_displacement <= -a:
        thumb_wheel_displacement += a
        return 1
    else:
        return False


def thumb_wheel_down(f, r, d, a):
    global thumb_wheel_displacement
    if f != _F.THUMB_WHEEL or r != 0:
        return False
    if a is None:
        return signed(d[0:2]) > 0 and signed(d[0:2])
    elif thumb_wheel_displacement >= a:
        thumb_wheel_displacement -= a
        return 1
    else:
        return False


def charging(f, r, d, a):
    if (f == _F.BATTERY_STATUS and r == 0 and 1 <= d[2] <= 4) or \
       (f == _F.BATTERY_VOLTAGE and r == 0 and d[2] & (1 << 7)) or \
       (f == _F.UNIFIED_BATTERY and r == 0 and 1 <= d[2] <= 3):
        return 1
    else:
        return False


TESTS = {
    'crown_right': [lambda f, r, d, a: f == _F.CROWN and r == 0 and d[1] < 128 and d[1], False],
    'crown_left': [lambda f, r, d, a: f == _F.CROWN and r == 0 and d[1] >= 128 and 256 - d[1], False],
    'crown_right_ratchet': [lambda f, r, d, a: f == _F.CROWN and r == 0 and d[2] < 128 and d[2], False],
    'crown_left_ratchet': [lambda f, r, d, a: f == _F.CROWN and r == 0 and d[2] >= 128 and 256 - d[2], False],
    'crown_tap': [lambda f, r, d, a: f == _F.CROWN and r == 0 and d[5] == 0x01 and d[5], False],
    'crown_start_press': [lambda f, r, d, a: f == _F.CROWN and r == 0 and d[6] == 0x01 and d[6], False],
    'crown_end_press': [lambda f, r, d, a: f == _F.CROWN and r == 0 and d[6] == 0x05 and d[6], False],
    'crown_pressed': [lambda f, r, d, a: f == _F.CROWN and r == 0 and d[6] >= 0x01 and d[6] <= 0x04 and d[6], False],
    'thumb_wheel_up': [thumb_wheel_up, True],
    'thumb_wheel_down': [thumb_wheel_down, True],
    'lowres_wheel_up': [lambda f, r, d, a: f == _F.LOWRES_WHEEL and r == 0 and signed(d[0:1]) > 0 and signed(d[0:1]), False],
    'lowres_wheel_down': [lambda f, r, d, a: f == _F.LOWRES_WHEEL and r == 0 and signed(d[0:1]) < 0 and signed(d[0:1]), False],
    'hires_wheel_up': [lambda f, r, d, a: f == _F.HIRES_WHEEL and r == 0 and signed(d[1:3]) > 0 and signed(d[1:3]), False],
    'hires_wheel_down': [lambda f, r, d, a: f == _F.HIRES_WHEEL and r == 0 and signed(d[1:3]) < 0 and signed(d[1:3]), False],
    'charging': [charging, False],
    'False': [lambda f, r, d, a: False, False],
    'True': [lambda f, r, d, a: True, False],
}

MOUSE_GESTURE_TESTS = {
    'mouse-down': ['Mouse Down'],
    'mouse-up': ['Mouse Up'],
    'mouse-left': ['Mouse Left'],
    'mouse-right': ['Mouse Right'],
    'mouse-noop': [],
}

COMPONENTS = {}


class RuleComponent:

    def compile(self, c):
        if isinstance(c, RuleComponent):
            return c
        elif isinstance(c, dict) and len(c) == 1:
            k, v = next(iter(c.items()))
            if k in COMPONENTS:
                return COMPONENTS[k](v)
        logger.warning('illegal component in rule: %s', c)
        return Condition()


class Rule(RuleComponent):

    def __init__(self, args, source=None, warn=True):
        self.components = [self.compile(a) for a in args]
        self.source = source

    def __str__(self):
        source = '(' + self.source + ')' if self.source else ''
        return 'Rule%s[%s]' % (source, ', '.join([c.__str__() for c in self.components]))

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate rule: %s', self)
        result = True
        for component in self.components:
            result = component.evaluate(feature, notification, device, status, result)
            if not isinstance(component, Action) and result is None:
                return None
            if isinstance(component, Condition) and not result:
                return result
        return result

    def once(self, feature, notification, device, status, last_result):
        self.evaluate(feature, notification, device, status, last_result)
        return False

    def data(self):
        return {'Rule': [c.data() for c in self.components]}


class Condition(RuleComponent):

    def __init__(self, *args):
        pass

    def __str__(self):
        return 'CONDITION'

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        return False


class Not(Condition):

    def __init__(self, op, warn=True):
        if isinstance(op, list) and len(op) == 1:
            op = op[0]
        self.op = op
        self.component = self.compile(op)

    def __str__(self):
        return 'Not: ' + str(self.component)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        result = self.component.evaluate(feature, notification, device, status, last_result)
        return None if result is None else not result

    def data(self):
        return {'Not': self.component.data()}


class Or(Condition):

    def __init__(self, args, warn=True):
        self.components = [self.compile(a) for a in args]

    def __str__(self):
        return 'Or: [' + ', '.join(str(c) for c in self.components) + ']'

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        result = False
        for component in self.components:
            result = component.evaluate(feature, notification, device, status, last_result)
            if not isinstance(component, Action) and result is None:
                return None
            if isinstance(component, Condition) and result:
                return result
        return result

    def data(self):
        return {'Or': [c.data() for c in self.components]}


class And(Condition):

    def __init__(self, args, warn=True):
        self.components = [self.compile(a) for a in args]

    def __str__(self):
        return 'And: [' + ', '.join(str(c) for c in self.components) + ']'

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        result = True
        for component in self.components:
            result = component.evaluate(feature, notification, device, status, last_result)
            if not isinstance(component, Action) and result is None:
                return None
            if isinstance(component, Condition) and not result:
                return result
        return result

    def data(self):
        return {'And': [c.data() for c in self.components]}


def x11_focus_prog():
    if not x11_setup():
        return None
    pid = wm_class = None
    window = xdisplay.get_input_focus().focus
    while window:
        pid = window.get_full_property(NET_WM_PID, 0)
        wm_class = window.get_wm_class()
        if wm_class and pid:
            break
        window = window.query_tree().parent
    try:
        name = psutil.Process(pid.value[0]).name() if pid else ''
    except Exception:
        name = ''
    return (wm_class[0], wm_class[1], name) if wm_class else (name, )


def x11_pointer_prog():
    if not x11_setup():
        return None
    pid = wm_class = None
    window = xdisplay.screen().root.query_pointer().child
    for child in reversed(window.query_tree().children):
        pid = child.get_full_property(NET_WM_PID, 0)
        wm_class = child.get_wm_class()
        if wm_class:
            break
    name = psutil.Process(pid.value[0]).name() if pid else ''
    return (wm_class[0], wm_class[1], name) if wm_class else (name, )


def gnome_dbus_focus_prog():
    if not gnome_dbus_interface_setup():
        return None
    wm_class = _dbus_interface.ActiveWindow()
    return (wm_class, ) if wm_class else None


def gnome_dbus_pointer_prog():
    if not gnome_dbus_interface_setup():
        return None
    wm_class = _dbus_interface.PointerOverWindow()
    return (wm_class, ) if wm_class else None


class Process(Condition):

    def __init__(self, process, warn=True):
        self.process = process
        if (not wayland and not x11_setup()) or (wayland and not gnome_dbus_interface_setup()):
            if warn:
                logger.warning(
                    'rules can only access active process in X11 or in Wayland under GNOME with Solaar Gnome extension - %s',
                    self
                )
        if not isinstance(process, str):
            if warn:
                logger.warning('rule Process argument not a string: %s', process)
            self.process = str(process)

    def __str__(self):
        return 'Process: ' + str(self.process)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        if not isinstance(self.process, str):
            return False
        focus = x11_focus_prog() if not wayland else gnome_dbus_focus_prog()
        result = any(bool(s and s.startswith(self.process)) for s in focus) if focus else None
        return result

    def data(self):
        return {'Process': str(self.process)}


class MouseProcess(Condition):

    def __init__(self, process, warn=True):
        self.process = process
        if (not wayland and not x11_setup()) or (wayland and not gnome_dbus_interface_setup()):
            if warn:
                logger.warning(
                    'rules cannot access active mouse process '
                    'in X11 or in Wayland under GNOME with Solaar Extension for GNOME - %s', self
                )
        if not isinstance(process, str):
            if warn:
                logger.warning('rule MouseProcess argument not a string: %s', process)
            self.process = str(process)

    def __str__(self):
        return 'MouseProcess: ' + str(self.process)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        if not isinstance(self.process, str):
            return False
        pointer_focus = x11_pointer_prog() if not wayland else gnome_dbus_pointer_prog()
        result = any(bool(s and s.startswith(self.process)) for s in pointer_focus) if pointer_focus else None
        return result

    def data(self):
        return {'MouseProcess': str(self.process)}


class Feature(Condition):

    def __init__(self, feature, warn=True):
        if not (isinstance(feature, str) and feature in _F):
            if warn:
                logger.warning('rule Feature argument not name of a feature: %s', feature)
            self.feature = None
        self.feature = _F[feature]

    def __str__(self):
        return 'Feature: ' + str(self.feature)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        return feature == self.feature

    def data(self):
        return {'Feature': str(self.feature)}


class Report(Condition):

    def __init__(self, report, warn=True):
        if not (isinstance(report, int)):
            if warn:
                logger.warning('rule Report argument not an integer: %s', report)
            self.report = -1
        else:
            self.report = report

    def __str__(self):
        return 'Report: ' + str(self.report)

    def evaluate(self, report, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        return (notification.address >> 4) == self.report

    def data(self):
        return {'Report': self.report}


# Setting(device, setting, [key], value...)
class Setting(Condition):

    def __init__(self, args, warn=True):
        if not (isinstance(args, list) and len(args) > 2):
            if warn:
                logger.warning('rule Setting argument not list with minimum length 3: %s', args)
            self.args = []
        else:
            self.args = args

    def __str__(self):
        return 'Setting: ' + ' '.join([str(a) for a in self.args])

    def evaluate(self, report, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        if len(self.args) < 3:
            return None
        dev = _Device.find(self.args[0]) if self.args[0] is not None else device
        if dev is None:
            logger.warning('Setting condition: device %s is not known', self.args[0])
            return False
        setting = next((s for s in dev.settings if s.name == self.args[1]), None)
        if setting is None:
            logger.warning('Setting condition: setting %s is not the name of a setting for %s', self.args[1], dev.name)
            return None
        # should the value argument be checked to be sure it is acceptable?? needs to be careful about boolean toggle
        # TODO add compare  methods for more validators
        try:
            result = setting.compare(self.args[2:], setting.read())
        except Exception as e:
            logger.warning('Setting condition: error when checking setting %s: %s', self.args, e)
            result = False
        return result

    def data(self):
        return {'Setting': self.args[:]}


MODIFIERS = {
    'Shift': int(Gdk.ModifierType.SHIFT_MASK),
    'Control': int(Gdk.ModifierType.CONTROL_MASK),
    'Alt': int(Gdk.ModifierType.MOD1_MASK),
    'Super': int(Gdk.ModifierType.MOD4_MASK)
}
MODIFIER_MASK = MODIFIERS['Shift'] + MODIFIERS['Control'] + MODIFIERS['Alt'] + MODIFIERS['Super']


class Modifiers(Condition):

    def __init__(self, modifiers, warn=True):
        modifiers = [modifiers] if isinstance(modifiers, str) else modifiers
        self.desired = 0
        self.modifiers = []
        for k in modifiers:
            if k in MODIFIERS:
                self.desired += MODIFIERS.get(k, 0)
                self.modifiers.append(k)
            else:
                if warn:
                    logger.warning('unknown rule Modifier value: %s', k)

    def __str__(self):
        return 'Modifiers: ' + str(self.desired)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        if gkeymap:
            current = gkeymap.get_modifier_state()  # get the current keyboard modifier
            return self.desired == (current & MODIFIER_MASK)
        else:
            logger.warning('no keymap so cannot determine modifier keys')
            return False

    def data(self):
        return {'Modifiers': [str(m) for m in self.modifiers]}


class Key(Condition):
    DOWN = 'pressed'
    UP = 'released'

    def __init__(self, args, warn=True):
        default_key = 0
        default_action = self.DOWN

        key, action = None, None

        if not args or not isinstance(args, (list, str)):
            if warn:
                logger.warning('rule Key arguments unknown: %s' % args)
            key = default_key
            action = default_action
        elif isinstance(args, str):
            logger.debug('rule Key assuming action "%s" for "%s"' % (default_action, args))
            key = args
            action = default_action
        elif isinstance(args, list):
            if len(args) == 1:
                logger.debug('rule Key assuming action "%s" for "%s"' % (default_action, args))
                key, action = args[0], default_action
            elif len(args) >= 2:
                key, action = args[:2]

        if isinstance(key, str) and key in _CONTROL:
            self.key = _CONTROL[key]
        else:
            if warn:
                logger.warning('rule Key key name not name of a Logitech key: %s' % key)
            self.key = default_key

        if isinstance(action, str) and action in (self.DOWN, self.UP):
            self.action = action
        else:
            if warn:
                logger.warning('rule Key action unknown: %s, assuming %s' % (action, default_action))
            self.action = default_action

    def __str__(self):
        return 'Key: %s (%s)' % ((str(self.key) if self.key else 'None'), self.action)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        return bool(self.key and self.key == (key_down if self.action == self.DOWN else key_up))

    def data(self):
        return {'Key': [str(self.key), self.action]}


class KeyIsDown(Condition):

    def __init__(self, args, warn=True):
        default_key = 0

        key = None

        if not args or not isinstance(args, str):
            if warn:
                logger.warning('rule KeyDown arguments unknown: %s' % args)
            key = default_key
        elif isinstance(args, str):
            key = args

        if isinstance(key, str) and key in _CONTROL:
            self.key = _CONTROL[key]
        else:
            if warn:
                logger.warning('rule Key key name not name of a Logitech key: %s' % key)
            self.key = default_key

    def __str__(self):
        return 'KeyIsDown: %s' % (str(self.key) if self.key else 'None')

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        return key_is_down(self.key)

    def data(self):
        return {'KeyIsDown': str(self.key)}


def bit_test(start, end, bits):
    return lambda f, r, d: int.from_bytes(d[start:end], byteorder='big', signed=True) & bits


def range_test(start, end, min, max):

    def range_test_helper(f, r, d):
        value = int.from_bytes(d[start:end], byteorder='big', signed=True)
        return min <= value <= max and (value if value else True)

    return range_test_helper


class Test(Condition):

    def __init__(self, test, warn=True):
        self.test = ''
        self.parameter = None
        if isinstance(test, str):
            test = [test]
        if isinstance(test, list) and all(isinstance(t, int) for t in test):
            if warn:
                logger.warning('Test rules consisting of numbers are deprecated, converting to a TestBytes condition')
            self.__class__ = TestBytes
            self.__init__(test, warn=warn)
        elif isinstance(test, list):
            if test[0] in MOUSE_GESTURE_TESTS:
                if warn:
                    logger.warning('mouse movement test %s deprecated, converting to a MouseGesture', test)
                self.__class__ = MouseGesture
                self.__init__(MOUSE_GESTURE_TESTS[0][test], warn=warn)
            elif test[0] in TESTS:
                self.test = test[0]
                self.function = TESTS[test[0]][0]
                self.parameter = test[1] if len(test) > 1 else None
            else:
                if warn:
                    logger.warning('rule Test name not name of a test: %s', test)
                self.test = 'False'
                self.function = TESTS['False'][0]
        else:
            if warn:
                logger.warning('rule Test argument not valid %s', test)

    def __str__(self):
        return 'Test: ' + str(self.test)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        return self.function(feature, notification.address, notification.data, self.parameter)

    def data(self):
        return {'Test': ([self.test, self.parameter] if self.parameter is not None else [self.test])}


class TestBytes(Condition):

    def __init__(self, test, warn=True):
        self.test = test
        if (
            isinstance(test, list) and 2 < len(test) <= 4 and all(isinstance(t, int) for t in test) and test[0] >= 0
            and test[0] <= 16 and test[1] >= 0 and test[1] <= 16 and test[0] < test[1]
        ):
            self.function = bit_test(*test) if len(test) == 3 else range_test(*test)
        else:
            if warn:
                logger.warning('rule TestBytes argument not valid %s', test)

    def __str__(self):
        return 'TestBytes: ' + str(self.test)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        return self.function(feature, notification.address, notification.data)

    def data(self):
        return {'TestBytes': self.test[:]}


class MouseGesture(Condition):
    MOVEMENTS = [
        'Mouse Up', 'Mouse Down', 'Mouse Left', 'Mouse Right', 'Mouse Up-left', 'Mouse Up-right', 'Mouse Down-left',
        'Mouse Down-right'
    ]

    def __init__(self, movements, warn=True):
        if isinstance(movements, str):
            movements = [movements]
        for x in movements:
            if x not in self.MOVEMENTS and x not in _CONTROL:
                if warn:
                    logger.warning('rule Mouse Gesture argument not direction or name of a Logitech key: %s', x)
        self.movements = movements

    def __str__(self):
        return 'MouseGesture: ' + ' '.join(self.movements)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        if feature == _F.MOUSE_GESTURE:
            d = notification.data
            data = _unpack('!' + (int(len(d) / 2) * 'h'), d)
            data_offset = 1
            movement_offset = 0
            if self.movements and self.movements[0] not in self.MOVEMENTS:  # matching against initiating key
                movement_offset = 1
                if self.movements[0] != str(_CONTROL[data[0]]):
                    return False
            for m in self.movements[movement_offset:]:
                if data_offset >= len(data):
                    return False
                if data[data_offset] == 0:
                    direction = xy_direction(data[data_offset + 1], data[data_offset + 2])
                    if m != direction:
                        return False
                    data_offset += 3
                elif data[data_offset] == 1:
                    if m != str(_CONTROL[data[data_offset + 1]]):
                        return False
                    data_offset += 2
            return data_offset == len(data)
        return False

    def data(self):
        return {'MouseGesture': [str(m) for m in self.movements]}


class Active(Condition):

    def __init__(self, devID, warn=True):
        if not (isinstance(devID, str)):
            if warn:
                logger.warning('rule Active argument not a string: %s', devID)
            self.devID = ''
        self.devID = devID

    def __str__(self):
        return 'Active: ' + str(self.devID)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        dev = _Device.find(self.devID)
        return bool(dev and dev.ping())

    def data(self):
        return {'Active': self.devID}


class Device(Condition):

    def __init__(self, devID, warn=True):
        if not (isinstance(devID, str)):
            if warn:
                logger.warning('rule Device argument not a string: %s', devID)
            self.devID = ''
        self.devID = devID

    def __str__(self):
        return 'Device: ' + str(self.devID)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        return device.unitId == self.devID or device.serial == self.devID

    def data(self):
        return {'Device': self.devID}


class Host(Condition):

    def __init__(self, host, warn=True):
        if not (isinstance(host, str)):
            if warn:
                logger.warning('rule Host Name argument not a string: %s', host)
            self.host = ''
        self.host = host

    def __str__(self):
        return 'Host: ' + str(self.host)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('evaluate condition: %s', self)
        import socket
        hostname = socket.getfqdn()
        return hostname.startswith(self.host)

    def data(self):
        return {'Host': self.host}


class Action(RuleComponent):

    def __init__(self, *args):
        pass

    def evaluate(self, feature, notification, device, status, last_result):
        return None


class KeyPress(Action):

    def __init__(self, args, warn=True):
        self.key_names, self.action = self.regularize_args(args)
        if not isinstance(self.key_names, list):
            if warn:
                logger.warning('rule KeyPress keys not key names %s', self.keys_names)
            self.key_symbols = []
        else:
            self.key_symbols = [XK_KEYS.get(k, None) for k in self.key_names]
        if not all(self.key_symbols):
            if warn:
                logger.warning('rule KeyPress keys not key names %s', self.key_names)
            self.key_symbols = []

    def regularize_args(self, args):
        action = CLICK
        if not isinstance(args, list):
            args = [args]
        keys = args
        if len(args) == 2 and args[1] in [CLICK, DEPRESS, RELEASE]:
            keys = [args[0]] if isinstance(args[0], str) else args[0]
            action = args[1]
        return keys, action

    # WARNING:  This is an attempt to reverse the keycode to keysym mappping in XKB.  It may not be completely general.
    def keysym_to_keycode(self, keysym, modifiers):  # maybe should take shift into account
        group = kbdgroup() or 0
        keycodes = gkeymap.get_entries_for_keyval(keysym)
        (keycode, level) = (None, None)
        for k in keycodes.keys:  # mappings that have the correct group
            if group == k.group and k.keycode < 256 and (level is None or k.level < level):
                keycode = k.keycode
                level = k.level
        if keycode or group == 0:
            return (keycode, level)
        for k in keycodes.keys:  # mappings for group 0 where keycode only has group 0 mappings
            if 0 == k.group and k.keycode < 256 and (level is None or k.level < level):
                (a, m, vs) = gkeymap.get_entries_for_keycode(k.keycode)
                if a and all(mk.group == 0 for mk in m):
                    keycode = k.keycode
                    level = k.level
        return (keycode, level)

    def __str__(self):
        return 'KeyPress: ' + ' '.join(self.key_names) + ' ' + self.action

    def needed(self, k, modifiers):
        code = modifier_code(k)
        return not (code is not None and modifiers & (1 << code))

    def mods(self, level, modifiers, direction):
        if level == 2 or level == 3:
            (sk, _) = self.keysym_to_keycode(XK_KEYS.get('ISO_Level3_Shift', None), modifiers)
            if sk and self.needed(sk, modifiers):
                simulate_key(sk, direction)
        if level == 1 or level == 3:
            (sk, _) = self.keysym_to_keycode(XK_KEYS.get('Shift_L', None), modifiers)
            if sk and self.needed(sk, modifiers):
                simulate_key(sk, direction)

    def keyDown(self, keysyms, modifiers):
        for k in keysyms:
            (keycode, level) = self.keysym_to_keycode(k, modifiers)
            if keycode is None:
                logger.warning('rule KeyPress key symbol not currently available %s', self)
            elif self.action != CLICK or self.needed(keycode, modifiers):  # only check needed when clicking
                self.mods(level, modifiers, _KEY_PRESS)
                simulate_key(keycode, _KEY_PRESS)

    def keyUp(self, keysyms, modifiers):
        for k in keysyms:
            (keycode, level) = self.keysym_to_keycode(k, modifiers)
            if keycode and (self.action != CLICK or self.needed(keycode, modifiers)):  # only check needed when clicking
                simulate_key(keycode, _KEY_RELEASE)
                self.mods(level, modifiers, _KEY_RELEASE)

    def evaluate(self, feature, notification, device, status, last_result):
        if gkeymap:
            current = gkeymap.get_modifier_state()
            if logger.isEnabledFor(logging.INFO):
                logger.info('KeyPress action: %s %s, group %s, modifiers %s', self.key_names, self.action, kbdgroup(), current)
            if self.action != RELEASE:
                self.keyDown(self.key_symbols, current)
            if self.action != DEPRESS:
                self.keyUp(reversed(self.key_symbols), current)
            _time.sleep(0.01)
        else:
            logger.warning('no keymap so cannot determine which keycode to send')
        return None

    def data(self):
        return {'KeyPress': [[str(k) for k in self.key_names], self.action]}


# KeyDown is dangerous as the key can auto-repeat and make your system unusable
# class KeyDown(KeyPress):
#    def evaluate(self, feature, notification, device, status, last_result):
#        super().keyDown(self.keys, current_key_modifiers)
# class KeyUp(KeyPress):
#    def evaluate(self, feature, notification, device, status, last_result):
#        super().keyUp(self.keys, current_key_modifiers)


class MouseScroll(Action):

    def __init__(self, amounts, warn=True):
        import numbers
        if len(amounts) == 1 and isinstance(amounts[0], list):
            amounts = amounts[0]
        if not (len(amounts) == 2 and all([isinstance(a, numbers.Number) for a in amounts])):
            if warn:
                logger.warning('rule MouseScroll argument not two numbers %s', amounts)
            amounts = [0, 0]
        self.amounts = amounts

    def __str__(self):
        return 'MouseScroll: ' + ' '.join([str(a) for a in self.amounts])

    def evaluate(self, feature, notification, device, status, last_result):
        import math
        import numbers
        amounts = self.amounts
        if isinstance(last_result, numbers.Number):
            amounts = [math.floor(last_result * a) for a in self.amounts]
        if logger.isEnabledFor(logging.INFO):
            logger.info('MouseScroll action: %s %s %s', self.amounts, last_result, amounts)
        dx, dy = amounts
        simulate_scroll(dx, dy)
        _time.sleep(0.01)
        return None

    def data(self):
        return {'MouseScroll': self.amounts[:]}


class MouseClick(Action):

    def __init__(self, args, warn=True):
        if len(args) == 1 and isinstance(args[0], list):
            args = args[0]
        if not isinstance(args, list):
            args = [args]
        self.button = str(args[0]) if len(args) >= 0 else None
        if self.button not in buttons:
            if warn:
                logger.warning('rule MouseClick action: button %s not known', self.button)
            self.button = None
        count = args[1] if len(args) >= 2 else 1
        try:
            self.count = int(count)
        except (ValueError, TypeError):
            if count in [CLICK, DEPRESS, RELEASE]:
                self.count = count
            elif warn:
                logger.warning('rule MouseClick action: argument %s should be an integer or CLICK, PRESS, or RELEASE', count)
                self.count = 1

    def __str__(self):
        return 'MouseClick: %s (%d)' % (self.button, self.count)

    def evaluate(self, feature, notification, device, status, last_result):
        if logger.isEnabledFor(logging.INFO):
            logger.info('MouseClick action: %d %s' % (self.count, self.button))
        if self.button and self.count:
            click(buttons[self.button], self.count)
        _time.sleep(0.01)
        return None

    def data(self):
        return {'MouseClick': [self.button, self.count]}


class Set(Action):

    def __init__(self, args, warn=True):
        if not (isinstance(args, list) and len(args) > 2):
            if warn:
                logger.warning('rule Set argument not list with minimum length 3: %s', args)
            self.args = []
        else:
            self.args = args

    def __str__(self):
        return 'Set: ' + ' '.join([str(a) for a in self.args])

    def evaluate(self, feature, notification, device, status, last_result):
        # importing here to avoid circular imports
        from solaar.ui.config_panel import change_setting as _change_setting

        if len(self.args) < 3:
            return None
        if logger.isEnabledFor(logging.INFO):
            logger.info('Set action: %s', self.args)
        dev = _Device.find(self.args[0]) if self.args[0] is not None else device
        if dev is None:
            logger.warning('Set action: device %s is not known', self.args[0])
            return None
        setting = next((s for s in dev.settings if s.name == self.args[1]), None)
        if setting is None:
            logger.warning('Set action: setting %s is not the name of a setting for %s', self.args[1], dev.name)
            return None
        args = setting.acceptable(self.args[2:], setting.read())
        if args is None:
            logger.warning('Set Action: invalid args %s for setting %s of %s', self.args[2:], self.args[1], self.args[0])
            return None
        _change_setting(dev, setting, args)
        return None

    def data(self):
        return {'Set': self.args[:]}


class Execute(Action):

    def __init__(self, args, warn=True):
        if isinstance(args, str):
            args = [args]
        if not (isinstance(args, list) and all(isinstance(arg), str) for arg in args):
            if warn:
                logger.warning('rule Execute argument not list of strings: %s', args)
            self.args = []
        else:
            self.args = args

    def __str__(self):
        return 'Execute: ' + ' '.join([a for a in self.args])

    def evaluate(self, feature, notification, device, status, last_result):
        import subprocess
        if logger.isEnabledFor(logging.INFO):
            logger.info('Execute action: %s', self.args)
        subprocess.Popen(self.args)
        return None

    def data(self):
        return {'Execute': self.args[:]}


class Later(Action):

    def __init__(self, args, warn=True):
        self.delay = 0
        self.rule = Rule([])
        self.components = []
        if not (isinstance(args, list)):
            args = [args]
        if not (isinstance(args, list) and len(args) >= 1):
            if warn:
                logger.warning('rule Later argument not list with minimum length 1: %s', args)
        elif not (isinstance(args[0], int)) or not 0 < args[0] < 101:
            if warn:
                logger.warning('rule Later argument delay not integer between 1 and 100: %s', args)
        else:
            self.delay = args[0]
            self.rule = Rule(args[1:], warn=warn)
            self.components = self.rule.components

    def __str__(self):
        return 'Later: [' + str(self.delay) + ', ' + ', '.join(str(c) for c in self.components) + ']'

    def evaluate(self, feature, notification, device, status, last_result):
        if self.delay and self.rule:
            GLib.timeout_add_seconds(self.delay, Rule.once, self.rule, feature, notification, device, status, last_result)
        return None

    def data(self):
        data = [c.data() for c in self.components]
        data.insert(0, self.delay)
        return {'Later': data}


COMPONENTS = {
    'Rule': Rule,
    'Not': Not,
    'Or': Or,
    'And': And,
    'Process': Process,
    'MouseProcess': MouseProcess,
    'Feature': Feature,
    'Report': Report,
    'Setting': Setting,
    'Modifiers': Modifiers,
    'Key': Key,
    'KeyIsDown': KeyIsDown,
    'Test': Test,
    'TestBytes': TestBytes,
    'MouseGesture': MouseGesture,
    'Active': Active,
    'Device': Device,
    'Host': Host,
    'KeyPress': KeyPress,
    'MouseScroll': MouseScroll,
    'MouseClick': MouseClick,
    'Set': Set,
    'Execute': Execute,
    'Later': Later,
}

built_in_rules = Rule([])
if True:
    built_in_rules = Rule([
        {'Rule': [  # Implement problematic keys for Craft and MX Master
            {'Rule': [{'Key': ['Brightness Down', 'pressed']}, {'KeyPress': 'XF86_MonBrightnessDown'}]},
            {'Rule': [{'Key': ['Brightness Up', 'pressed']}, {'KeyPress': 'XF86_MonBrightnessUp'}]},
        ]},
        # {'Rule': [  # In firefox, crown emits keys that move up and down if not pressed, rotate through tabs otherwise
        #     {'Process': 'firefox'},
        #     {'Rule': [{'Test': 'crown_pressed'}, {'Test': 'crown_right_ratchet'}, {'KeyPress': ['Control_R', 'Tab']}]},
        #     {'Rule': [{'Test': 'crown_pressed'},
        #               {'Test': 'crown_left_ratchet'},
        #               {'KeyPress': ['Control_R', 'Shift_R', 'Tab']}]},
        #     {'Rule': [{'Test': 'crown_right_ratchet'}, {'KeyPress': 'Down'}]},
        #     {'Rule': [{'Test': 'crown_left_ratchet'}, {'KeyPress': 'Up'}]},
        # ]},
        # {'Rule': [  # Otherwise, crown movements emit keys that modify volume if not pressed, move between tracks otherwise
        #     {'Feature': 'CROWN'}, {'Report': 0x0},
        #     {'Rule': [{'Test': 'crown_pressed'}, {'Test': 'crown_right_ratchet'}, {'KeyPress': 'XF86_AudioNext'}]},
        #     {'Rule': [{'Test': 'crown_pressed'}, {'Test': 'crown_left_ratchet'}, {'KeyPress': 'XF86_AudioPrev'}]},
        #     {'Rule': [{'Test': 'crown_right_ratchet'}, {'KeyPress': 'XF86_AudioRaiseVolume'}]},
        #     {'Rule': [{'Test': 'crown_left_ratchet'}, {'KeyPress': 'XF86_AudioLowerVolume'}]}
        # ]},
    ])

keys_down = []
g_keys_down = 0
m_keys_down = 0
mr_key_down = False
thumb_wheel_displacement = 0


def key_is_down(key):
    if key == _CONTROL.MR:
        return mr_key_down
    elif _CONTROL.M1 <= key <= _CONTROL.M8:
        return bool(m_keys_down & (0x01 << (key - _CONTROL.M1)))
    elif _CONTROL.G1 <= key <= _CONTROL.G32:
        return bool(g_keys_down & (0x01 << (key - _CONTROL.G1)))
    else:
        return key in keys_down


def evaluate_rules(feature, notification, device, status):
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('evaluating rules on %s', notification)
    rules.evaluate(feature, notification, device, status, True)


# process a notification
def process_notification(device, status, notification, feature):
    global keys_down, g_keys_down, m_keys_down, mr_key_down, key_down, key_up, thumb_wheel_displacement
    key_down, key_up = None, None
    # need to keep track of keys that are down to find a new key down
    if feature == _F.REPROG_CONTROLS_V4 and notification.address == 0x00:
        new_keys_down = _unpack('!4H', notification.data[:8])
        for key in new_keys_down:
            if key and key not in keys_down:
                key_down = key
        for key in keys_down:
            if key and key not in new_keys_down:
                key_up = key
        keys_down = new_keys_down
    # and also G keys down
    elif feature == _F.GKEY and notification.address == 0x00:
        new_g_keys_down = _unpack('<I', notification.data[:4])[0]
        for i in range(32):
            if new_g_keys_down & (0x01 << i) and not g_keys_down & (0x01 << i):
                key_down = _CONTROL['G' + str(i + 1)]
            if g_keys_down & (0x01 << i) and not new_g_keys_down & (0x01 << i):
                key_up = _CONTROL['G' + str(i + 1)]
        g_keys_down = new_g_keys_down
    # and also M keys down
    elif feature == _F.MKEYS and notification.address == 0x00:
        new_m_keys_down = _unpack('!1B', notification.data[:1])[0]
        for i in range(1, 9):
            if new_m_keys_down & (0x01 << (i - 1)) and not m_keys_down & (0x01 << (i - 1)):
                key_down = _CONTROL['M' + str(i)]
            if m_keys_down & (0x01 << (i - 1)) and not new_m_keys_down & (0x01 << (i - 1)):
                key_up = _CONTROL['M' + str(i)]
        m_keys_down = new_m_keys_down
    # and also MR key
    elif feature == _F.MR and notification.address == 0x00:
        new_mr_key_down = _unpack('!1B', notification.data[:1])[0]
        if not mr_key_down and new_mr_key_down:
            key_down = _CONTROL['MR']
        if mr_key_down and not new_mr_key_down:
            key_up = _CONTROL['MR']
        mr_key_down = new_mr_key_down
    # keep track of thumb wheel movment
    elif feature == _F.THUMB_WHEEL and notification.address == 0x00:
        if notification.data[4] <= 0x01:  # when wheel starts, zero out last movement
            thumb_wheel_displacement = 0
        thumb_wheel_displacement += signed(notification.data[0:2])

    GLib.idle_add(evaluate_rules, feature, notification, device, status)


_XDG_CONFIG_HOME = _os.environ.get('XDG_CONFIG_HOME') or _path.expanduser(_path.join('~', '.config'))
_file_path = _path.join(_XDG_CONFIG_HOME, 'solaar', 'rules.yaml')

rules = built_in_rules


def _save_config_rule_file(file_name=_file_path):
    # This is a trick to show str/float/int lists in-line (inspired by https://stackoverflow.com/a/14001707)
    class inline_list(list):
        pass

    def blockseq_rep(dumper, data):
        return dumper.represent_sequence('tag:yaml.org,2002:seq', data, flow_style=True)

    _yaml_add_representer(inline_list, blockseq_rep)

    def convert(elem):
        if isinstance(elem, list):
            if len(elem) == 1 and isinstance(elem[0], (int, str, float)):
                # All diversion classes that expect a list of scalars also support a single scalar without a list
                return elem[0]
            if all(isinstance(c, (int, str, float)) for c in elem):
                return inline_list([convert(c) for c in elem])
            return [convert(c) for c in elem]
        if isinstance(elem, dict):
            return {k: convert(v) for k, v in elem.items()}
        if isinstance(elem, NamedInt):
            return int(elem)
        return elem

    # YAML format settings
    dump_settings = {
        'encoding': 'utf-8',
        'explicit_start': True,
        'explicit_end': True,
        'default_flow_style': False
        # 'version': (1, 3),  # it would be printed for every rule
    }
    # Save only user-defined rules
    rules_to_save = sum((r.data()['Rule'] for r in rules.components if r.source == file_name), [])
    if True:  # save even if there are no rules to save
        if logger.isEnabledFor(logging.INFO):
            logger.info('saving %d rule(s) to %s', len(rules_to_save), file_name)
        try:
            with open(file_name, 'w') as f:
                if rules_to_save:
                    f.write('%YAML 1.3\n')  # Write version manually
                _yaml_dump_all(convert([r['Rule'] for r in rules_to_save]), f, **dump_settings)
        except Exception as e:
            logger.error('failed to save to %s\n%s', file_name, e)
            return False
    return True


def _load_config_rule_file():
    global rules
    loaded_rules = []
    if _path.isfile(_file_path):
        try:
            with open(_file_path) as config_file:
                loaded_rules = []
                for loaded_rule in _yaml_safe_load_all(config_file):
                    rule = Rule(loaded_rule, source=_file_path)
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug('load rule: %s', rule)
                    loaded_rules.append(rule)
                if logger.isEnabledFor(logging.INFO):
                    logger.info('loaded %d rules from %s', len(loaded_rules), config_file.name)
        except Exception as e:
            logger.error('failed to load from %s\n%s', _file_path, e)
    rules = Rule([Rule(loaded_rules, source=_file_path), built_in_rules])


_load_config_rule_file()



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Low-level interface for devices connected through a Logitech Universal
Receiver (UR).

Uses the HID api exposed through hidapi.py, a Python thin layer over a native
implementation.

Incomplete. Based on a bit of documentation, trial-and-error, and guesswork.

References:
http://julien.danjou.info/blog/2012/logitech-k750-linux-support
http://6xq.net/git/lars/lshidpp.git/plain/doc/
"""

import logging

from . import listener, status  # noqa: F401
from .base import DeviceUnreachable, NoReceiver, NoSuchDevice  # noqa: F401
from .common import strhex  # noqa: F401
from .device import Device  # noqa: F401
from .hidpp20 import FeatureCallError, FeatureNotSupported  # noqa: F401
from .receiver import Receiver  # noqa: F401

logger = logging.getLogger(__name__)
logger.setLevel(logging.root.level)
# if logging.root.level > logging.DEBUG:
#     logger.addHandler(logging.NullHandler())
#     logger.propagate = 0

del logging

__version__ = '0.9'



## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Handles incoming events from the receiver/devices, updating the related
# status object as appropriate.

import logging
import threading as _threading

from struct import unpack as _unpack

from . import diversion as _diversion
from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20
from . import settings_templates as _st
from .base import DJ_MESSAGE_ID as _DJ_MESSAGE_ID
from .common import strhex as _strhex
from .i18n import _
from .status import ALERT as _ALERT
from .status import KEYS as _K

logger = logging.getLogger(__name__)

_R = _hidpp10.REGISTERS
_F = _hidpp20.FEATURE

#
#
#

notification_lock = _threading.Lock()


def process(device, notification):
    assert device
    assert notification

    assert hasattr(device, 'status')
    status = device.status
    assert status is not None

    if not device.isDevice:
        return _process_receiver_notification(device, status, notification)

    return _process_device_notification(device, status, notification)


#
#
#


def _process_receiver_notification(receiver, status, n):
    # supposedly only 0x4x notifications arrive for the receiver
    assert n.sub_id & 0x40 == 0x40

    if n.sub_id == 0x4A:  # pairing lock notification
        status.lock_open = bool(n.address & 0x01)
        reason = (_('pairing lock is open') if status.lock_open else _('pairing lock is closed'))
        if logger.isEnabledFor(logging.INFO):
            logger.info('%s: %s', receiver, reason)
        status[_K.ERROR] = None
        if status.lock_open:
            status.new_device = None
        pair_error = ord(n.data[:1])
        if pair_error:
            status[_K.ERROR] = error_string = _hidpp10.PAIRING_ERRORS[pair_error]
            status.new_device = None
            logger.warning('pairing error %d: %s', pair_error, error_string)
        status.changed(reason=reason)
        return True

    elif n.sub_id == _R.discovery_status_notification:  # Bolt pairing
        with notification_lock:
            status.discovering = n.address == 0x00
            reason = (_('discovery lock is open') if status.discovering else _('discovery lock is closed'))
            if logger.isEnabledFor(logging.INFO):
                logger.info('%s: %s', receiver, reason)
            status[_K.ERROR] = None
            if status.discovering:
                status.counter = status.device_address = status.device_authentication = status.device_name = None
            status.device_passkey = None
            discover_error = ord(n.data[:1])
            if discover_error:
                status[_K.ERROR] = discover_string = _hidpp10.BOLT_PAIRING_ERRORS[discover_error]
                logger.warning('bolt discovering error %d: %s', discover_error, discover_string)
            status.changed(reason=reason)
            return True

    elif n.sub_id == _R.device_discovery_notification:  # Bolt pairing
        with notification_lock:
            counter = n.address + n.data[0] * 256  # notification counter
            if status.counter is None:
                status.counter = counter
            else:
                if not status.counter == counter:
                    return None
            if n.data[1] == 0:
                status.device_kind = n.data[3]
                status.device_address = n.data[6:12]
                status.device_authentication = n.data[14]
            elif n.data[1] == 1:
                status.device_name = n.data[3:3 + n.data[2]].decode('utf-8')
            return True

    elif n.sub_id == _R.pairing_status_notification:  # Bolt pairing
        with notification_lock:
            status.device_passkey = None
            status.lock_open = n.address == 0x00
            reason = (_('pairing lock is open') if status.lock_open else _('pairing lock is closed'))
            if logger.isEnabledFor(logging.INFO):
                logger.info('%s: %s', receiver, reason)
            status[_K.ERROR] = None
            if not status.lock_open:
                status.counter = status.device_address = status.device_authentication = status.device_name = None
            pair_error = n.data[0]
            if status.lock_open:
                status.new_device = None
            elif n.address == 0x02 and not pair_error:
                status.new_device = receiver.register_new_device(n.data[7])
            if pair_error:
                status[_K.ERROR] = error_string = _hidpp10.BOLT_PAIRING_ERRORS[pair_error]
                status.new_device = None
                logger.warning('pairing error %d: %s', pair_error, error_string)
            status.changed(reason=reason)
            return True

    elif n.sub_id == _R.passkey_request_notification:  # Bolt pairing
        with notification_lock:
            status.device_passkey = n.data[0:6].decode('utf-8')
            return True

    elif n.sub_id == _R.passkey_pressed_notification:  # Bolt pairing
        return True

    logger.warning('%s: unhandled notification %s', receiver, n)


#
#
#


def _process_device_notification(device, status, n):
    # incoming packets with SubId >= 0x80 are supposedly replies from
    # HID++ 1.0 requests, should never get here
    assert n.sub_id & 0x80 == 0

    if n.sub_id == 00:  # no-op feature notification, dispose of it quickly
        return False

    # Allow the device object to handle the notification using custom
    # per-device state.
    handling_ret = device.handle_notification(n)
    if handling_ret is not None:
        return handling_ret

    # 0x40 to 0x7F appear to be HID++ 1.0 or DJ notifications
    if n.sub_id >= 0x40:
        if n.report_id == _DJ_MESSAGE_ID:
            return _process_dj_notification(device, status, n)
        else:
            return _process_hidpp10_notification(device, status, n)

    # These notifications are from the device itself, so it must be active
    device.online = True
    # At this point, we need to know the device's protocol, otherwise it's
    # possible to not know how to handle it.
    assert device.protocol is not None

    # some custom battery events for HID++ 1.0 devices
    if device.protocol < 2.0:
        return _process_hidpp10_custom_notification(device, status, n)

    # assuming 0x00 to 0x3F are feature (HID++ 2.0) notifications
    if not device.features:
        logger.warning('%s: feature notification but features not set up: %02X %s', device, n.sub_id, n)
        return False
    try:
        feature = device.features.get_feature(n.sub_id)
    except IndexError:
        logger.warning('%s: notification from invalid feature index %02X: %s', device, n.sub_id, n)
        return False

    return _process_feature_notification(device, status, n, feature)


def _process_dj_notification(device, status, n):
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('%s (%s) DJ %s', device, device.protocol, n)

    if n.sub_id == 0x40:
        # do all DJ paired notifications also show up as HID++ 1.0 notifications?
        if logger.isEnabledFor(logging.INFO):
            logger.info('%s: ignoring DJ unpaired: %s', device, n)
        return True

    if n.sub_id == 0x41:
        # do all DJ paired notifications also show up as HID++ 1.0 notifications?
        if logger.isEnabledFor(logging.INFO):
            logger.info('%s: ignoring DJ paired: %s', device, n)
        return True

    if n.sub_id == 0x42:
        connected = not n.address & 0x01
        if logger.isEnabledFor(logging.INFO):
            logger.info('%s: DJ connection: %s %s', device, connected, n)
        status.changed(active=connected, alert=_ALERT.NONE, reason=_('connected') if connected else _('disconnected'))
        return True

    logger.warning('%s: unrecognized DJ %s', device, n)


def _process_hidpp10_custom_notification(device, status, n):
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('%s (%s) custom notification %s', device, device.protocol, n)

    if n.sub_id in (_R.battery_status, _R.battery_charge):
        # message layout: 10 ix <register> <xx> <yy> <zz> <00>
        assert n.data[-1:] == b'\x00'
        data = chr(n.address).encode() + n.data
        charge, next_charge, status_text, voltage = _hidpp10.parse_battery_status(n.sub_id, data)
        status.set_battery_info(charge, next_charge, status_text, voltage)
        return True

    if n.sub_id == _R.keyboard_illumination:
        # message layout: 10 ix 17("address")  <??> <?> <??> <light level 1=off..5=max>
        # TODO anything we can do with this?
        if logger.isEnabledFor(logging.INFO):
            logger.info('illumination event: %s', n)
        return True

    logger.warning('%s: unrecognized %s', device, n)


def _process_hidpp10_notification(device, status, n):
    # device unpairing
    if n.sub_id == 0x40:
        if n.address == 0x02:
            # device un-paired
            status.clear()
            device.wpid = None
            device.status = None
            if device.number in device.receiver:
                del device.receiver[device.number]
            status.changed(active=False, alert=_ALERT.ALL, reason=_('unpaired'))
        else:
            logger.warning('%s: disconnection with unknown type %02X: %s', device, n.address, n)
        return True

    # device connection (and disconnection)
    if n.sub_id == 0x41:
        flags = ord(n.data[:1]) & 0xF0
        if n.address == 0x02:  # very old 27 MHz protocol
            wpid = '00' + _strhex(n.data[2:3])
            link_established = True
            link_encrypted = bool(flags & 0x80)
        elif n.address > 0x00:  # all other protocols are supposed to be almost the same
            wpid = _strhex(n.data[2:3] + n.data[1:2])
            link_established = not (flags & 0x40)
            link_encrypted = bool(flags & 0x20) or n.address == 0x10  # Bolt protocol always encrypted
        else:
            logger.warning('%s: connection notification with unknown protocol %02X: %s', device.number, n.address, n)
            return True
        if wpid != device.wpid:
            logger.warning('%s wpid mismatch, got %s', device, wpid)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                '%s: protocol %s connection notification: software=%s, encrypted=%s, link=%s, payload=%s', device, n.address,
                bool(flags & 0x10), link_encrypted, link_established, bool(flags & 0x80)
            )
        status[_K.LINK_ENCRYPTED] = link_encrypted
        status.changed(active=link_established)
        return True

    if n.sub_id == 0x49:
        # raw input event? just ignore it
        # if n.address == 0x01, no idea what it is, but they keep on coming
        # if n.address == 0x03, appears to be an actual input event,
        #     because they only come when input happents
        return True

    # power notification
    if n.sub_id == 0x4B:
        if n.address == 0x01:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('%s: device powered on', device)
            reason = status.to_string() or _('powered on')
            status.changed(active=True, alert=_ALERT.NOTIFICATION, reason=reason)
        else:
            logger.warning('%s: unknown %s', device, n)
        return True

    logger.warning('%s: unrecognized %s', device, n)


def _process_feature_notification(device, status, n, feature):
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('%s: notification for feature %s, report %s, data %s', device, feature, n.address >> 4, _strhex(n.data))

    if feature == _F.BATTERY_STATUS:
        if n.address == 0x00:
            _ignore, discharge_level, discharge_next_level, battery_status, voltage = _hidpp20.decipher_battery_status(n.data)
            status.set_battery_info(discharge_level, discharge_next_level, battery_status, voltage)
        elif n.address == 0x10:
            if logger.isEnabledFor(logging.INFO):
                logger.info('%s: spurious BATTERY status %s', device, n)
        else:
            logger.warning('%s: unknown BATTERY %s', device, n)

    elif feature == _F.BATTERY_VOLTAGE:
        if n.address == 0x00:
            _ignore, level, nextl, battery_status, voltage = _hidpp20.decipher_battery_voltage(n.data)
            status.set_battery_info(level, nextl, battery_status, voltage)
        else:
            logger.warning('%s: unknown VOLTAGE %s', device, n)

    elif feature == _F.UNIFIED_BATTERY:
        if n.address == 0x00:
            _ignore, level, nextl, battery_status, voltage = _hidpp20.decipher_battery_unified(n.data)
            status.set_battery_info(level, nextl, battery_status, voltage)
        else:
            logger.warning('%s: unknown UNIFIED BATTERY %s', device, n)

    elif feature == _F.ADC_MEASUREMENT:
        if n.address == 0x00:
            result = _hidpp20.decipher_adc_measurement(n.data)
            if result:
                _ignore, level, nextl, battery_status, voltage = result
                status.set_battery_info(level, nextl, battery_status, voltage)
            else:  # this feature is used to signal device becoming inactive
                status.changed(active=False)
        else:
            logger.warning('%s: unknown ADC MEASUREMENT %s', device, n)

    elif feature == _F.SOLAR_DASHBOARD:
        if n.data[5:9] == b'GOOD':
            charge, lux, adc = _unpack('!BHH', n.data[:5])
            # guesstimate the battery voltage, emphasis on 'guess'
            # status_text = '%1.2fV' % (adc * 2.67793237653 / 0x0672)
            status_text = _hidpp20.BATTERY_STATUS.discharging
            if n.address == 0x00:
                status[_K.LIGHT_LEVEL] = None
                status.set_battery_info(charge, None, status_text, None)
            elif n.address == 0x10:
                status[_K.LIGHT_LEVEL] = lux
                if lux > 200:
                    status_text = _hidpp20.BATTERY_STATUS.recharging
                status.set_battery_info(charge, None, status_text, None)
            elif n.address == 0x20:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug('%s: Light Check button pressed', device)
                status.changed(alert=_ALERT.SHOW_WINDOW)
                # first cancel any reporting
                # device.feature_request(_F.SOLAR_DASHBOARD)
                # trigger a new report chain
                reports_count = 15
                reports_period = 2  # seconds
                device.feature_request(_F.SOLAR_DASHBOARD, 0x00, reports_count, reports_period)
            else:
                logger.warning('%s: unknown SOLAR CHARGE %s', device, n)
        else:
            logger.warning('%s: SOLAR CHARGE not GOOD? %s', device, n)

    elif feature == _F.WIRELESS_DEVICE_STATUS:
        if n.address == 0x00:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('wireless status: %s', n)
            reason = 'powered on' if n.data[2] == 1 else None
            if n.data[1] == 1:  # device is asking for software reconfiguration so need to change status
                alert = _ALERT.NONE
                status.changed(active=True, alert=alert, reason=reason, push=True)
        else:
            logger.warning('%s: unknown WIRELESS %s', device, n)

    elif feature == _F.TOUCHMOUSE_RAW_POINTS:
        if n.address == 0x00:
            if logger.isEnabledFor(logging.INFO):
                logger.info('%s: TOUCH MOUSE points %s', device, n)
        elif n.address == 0x10:
            touch = ord(n.data[:1])
            button_down = bool(touch & 0x02)
            mouse_lifted = bool(touch & 0x01)
            if logger.isEnabledFor(logging.INFO):
                logger.info('%s: TOUCH MOUSE status: button_down=%s mouse_lifted=%s', device, button_down, mouse_lifted)
        else:
            logger.warning('%s: unknown TOUCH MOUSE %s', device, n)

    # TODO: what are REPROG_CONTROLS_V{2,3}?
    elif feature == _F.REPROG_CONTROLS:
        if n.address == 0x00:
            if logger.isEnabledFor(logging.INFO):
                logger.info('%s: reprogrammable key: %s', device, n)
        else:
            logger.warning('%s: unknown REPROG_CONTROLS %s', device, n)

    elif feature == _F.BACKLIGHT2:
        if (n.address == 0x00):
            level = _unpack('!B', n.data[1:2])[0]
            from solaar.ui.config_panel import record_setting  # prevent circular import
            record_setting(device, _st.Backlight2Level, [level])

    elif feature == _F.REPROG_CONTROLS_V4:
        if n.address == 0x00:
            if logger.isEnabledFor(logging.DEBUG):
                cid1, cid2, cid3, cid4 = _unpack('!HHHH', n.data[:8])
                logger.debug('%s: diverted controls pressed: 0x%x, 0x%x, 0x%x, 0x%x', device, cid1, cid2, cid3, cid4)
        elif n.address == 0x10:
            if logger.isEnabledFor(logging.DEBUG):
                dx, dy = _unpack('!hh', n.data[:4])
                logger.debug('%s: rawXY dx=%i dy=%i', device, dx, dy)
        elif n.address == 0x20:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('%s: received analyticsKeyEvents', device)
        elif logger.isEnabledFor(logging.INFO):
            logger.info('%s: unknown REPROG_CONTROLS_V4 %s', device, n)

    elif feature == _F.HIRES_WHEEL:
        if (n.address == 0x00):
            if logger.isEnabledFor(logging.INFO):
                flags, delta_v = _unpack('>bh', n.data[:3])
                high_res = (flags & 0x10) != 0
                periods = flags & 0x0f
                logger.info('%s: WHEEL: res: %d periods: %d delta V:%-3d', device, high_res, periods, delta_v)
        elif (n.address == 0x10):
            ratchet = n.data[0]
            if logger.isEnabledFor(logging.INFO):
                logger.info('%s: WHEEL: ratchet: %d', device, ratchet)
            if ratchet < 2:  # don't process messages with unusual ratchet values
                from solaar.ui.config_panel import record_setting  # prevent circular import
                record_setting(device, _st.ScrollRatchet, [2 if ratchet else 1])
        else:
            if logger.isEnabledFor(logging.INFO):
                logger.info('%s: unknown WHEEL %s', device, n)

    elif feature == _F.ONBOARD_PROFILES:
        if (n.address > 0x10):
            if logger.isEnabledFor(logging.INFO):
                logger.info('%s: unknown ONBOARD PROFILES %s', device, n)
        else:
            if (n.address == 0x00):
                profile_sector = _unpack('!H', n.data[:2])[0]
                if profile_sector:
                    profile_change(device, profile_sector)
            elif (n.address == 0x10):
                resolution_index = _unpack('!B', n.data[:1])[0]
                profile_sector = _unpack('!H', device.feature_request(_F.ONBOARD_PROFILES, 0x40)[:2])[0]
                for profile in device.profiles.profiles.values() if device.profiles else []:
                    if profile.sector == profile_sector:
                        from solaar.ui.config_panel import record_setting  # prevent circular import
                        record_setting(device, _st.AdjustableDpi, [profile.resolutions[resolution_index]])

    _diversion.process_notification(device, status, n, feature)
    return True


# change UI to show result of onboard profile change
def profile_change(device, profile_sector):
    from solaar.ui.config_panel import record_setting  # prevent circular import
    record_setting(device, _st.OnboardProfiles, [profile_sector])
    for profile in device.profiles.profiles.values() if device.profiles else []:
        if profile.sector == profile_sector:
            resolution_index = profile.resolution_default_index
            record_setting(device, _st.AdjustableDpi, [profile.resolutions[resolution_index]])
            record_setting(device, _st.ReportRate, [profile.report_rate])
            break



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Generic Human Interface Device API.

This provides a python interface to libusb's hidapi library which,
unlike udev, is available for non-linux platforms.
See https://github.com/libusb/hidapi for how to obtain binaries.

Parts of this code are adapted from https://github.com/apmorton/pyhidapi
which is MIT licensed.
"""
import atexit
import ctypes
import logging
import platform as _platform

from collections import namedtuple
from threading import Thread
from time import sleep

logger = logging.getLogger(__name__)

native_implementation = 'hidapi'

# Device info as expected by Solaar
DeviceInfo = namedtuple(
    'DeviceInfo', [
        'path',
        'bus_id',
        'vendor_id',
        'product_id',
        'interface',
        'driver',
        'manufacturer',
        'product',
        'serial',
        'release',
        'isDevice',
        'hidpp_short',
        'hidpp_long',
    ]
)
del namedtuple

# Global handle to hidapi
_hidapi = None

# hidapi binary names for various platforms
_library_paths = (
    'libhidapi-hidraw.so', 'libhidapi-hidraw.so.0', 'libhidapi-libusb.so', 'libhidapi-libusb.so.0',
    'libhidapi-iohidmanager.so', 'libhidapi-iohidmanager.so.0', 'libhidapi.dylib', 'hidapi.dll', 'libhidapi-0.dll'
)

for lib in _library_paths:
    try:
        _hidapi = ctypes.cdll.LoadLibrary(lib)
        break
    except OSError:
        pass
else:
    raise ImportError(f"Unable to load hdiapi library, tried: {' '.join(_library_paths)}")


# Retrieve version of hdiapi library
class _cHidApiVersion(ctypes.Structure):
    _fields_ = [
        ('major', ctypes.c_int),
        ('minor', ctypes.c_int),
        ('patch', ctypes.c_int),
    ]


_hidapi.hid_version.argtypes = []
_hidapi.hid_version.restype = ctypes.POINTER(_cHidApiVersion)
_hid_version = _hidapi.hid_version()


# Construct device info struct based on API version
class _cDeviceInfo(ctypes.Structure):

    def as_dict(self):
        return {name: getattr(self, name) for name, _t in self._fields_ if name != 'next'}


# Low level hdiapi device info struct
# See https://github.com/libusb/hidapi/blob/master/hidapi/hidapi.h#L143
_cDeviceInfo_fields = [
    ('path', ctypes.c_char_p),
    ('vendor_id', ctypes.c_ushort),
    ('product_id', ctypes.c_ushort),
    ('serial_number', ctypes.c_wchar_p),
    ('release_number', ctypes.c_ushort),
    ('manufacturer_string', ctypes.c_wchar_p),
    ('product_string', ctypes.c_wchar_p),
    ('usage_page', ctypes.c_ushort),
    ('usage', ctypes.c_ushort),
    ('interface_number', ctypes.c_int),
    ('next', ctypes.POINTER(_cDeviceInfo)),
]
if _hid_version.contents.major >= 0 and _hid_version.contents.minor >= 13:
    _cDeviceInfo_fields.append(('bus_type', ctypes.c_int))
_cDeviceInfo._fields_ = _cDeviceInfo_fields

# Set up hidapi functions
_hidapi.hid_init.argtypes = []
_hidapi.hid_init.restype = ctypes.c_int
_hidapi.hid_exit.argtypes = []
_hidapi.hid_exit.restype = ctypes.c_int
_hidapi.hid_enumerate.argtypes = [ctypes.c_ushort, ctypes.c_ushort]
_hidapi.hid_enumerate.restype = ctypes.POINTER(_cDeviceInfo)
_hidapi.hid_free_enumeration.argtypes = [ctypes.POINTER(_cDeviceInfo)]
_hidapi.hid_free_enumeration.restype = None
_hidapi.hid_open.argtypes = [ctypes.c_ushort, ctypes.c_ushort, ctypes.c_wchar_p]
_hidapi.hid_open.restype = ctypes.c_void_p
_hidapi.hid_open_path.argtypes = [ctypes.c_char_p]
_hidapi.hid_open_path.restype = ctypes.c_void_p
_hidapi.hid_write.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_size_t]
_hidapi.hid_write.restype = ctypes.c_int
_hidapi.hid_read_timeout.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_size_t, ctypes.c_int]
_hidapi.hid_read_timeout.restype = ctypes.c_int
_hidapi.hid_read.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_size_t]
_hidapi.hid_read.restype = ctypes.c_int
_hidapi.hid_get_input_report.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_size_t]
_hidapi.hid_get_input_report.restype = ctypes.c_int
_hidapi.hid_set_nonblocking.argtypes = [ctypes.c_void_p, ctypes.c_int]
_hidapi.hid_set_nonblocking.restype = ctypes.c_int
_hidapi.hid_send_feature_report.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_int]
_hidapi.hid_send_feature_report.restype = ctypes.c_int
_hidapi.hid_get_feature_report.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_size_t]
_hidapi.hid_get_feature_report.restype = ctypes.c_int
_hidapi.hid_close.argtypes = [ctypes.c_void_p]
_hidapi.hid_close.restype = None
_hidapi.hid_get_manufacturer_string.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p, ctypes.c_size_t]
_hidapi.hid_get_manufacturer_string.restype = ctypes.c_int
_hidapi.hid_get_product_string.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p, ctypes.c_size_t]
_hidapi.hid_get_product_string.restype = ctypes.c_int
_hidapi.hid_get_serial_number_string.argtypes = [ctypes.c_void_p, ctypes.c_wchar_p, ctypes.c_size_t]
_hidapi.hid_get_serial_number_string.restype = ctypes.c_int
_hidapi.hid_get_indexed_string.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_wchar_p, ctypes.c_size_t]
_hidapi.hid_get_indexed_string.restype = ctypes.c_int
_hidapi.hid_error.argtypes = [ctypes.c_void_p]
_hidapi.hid_error.restype = ctypes.c_wchar_p

# Initialize hidapi
_hidapi.hid_init()
atexit.register(_hidapi.hid_exit)

# Solaar opens the same device more than once which will fail unless we
# allow non-exclusive opening. On windows opening with shared access is
# the default, for macOS we need to set it explicitly.
if _platform.system() == 'Darwin':
    _hidapi.hid_darwin_set_open_exclusive.argtypes = [ctypes.c_int]
    _hidapi.hid_darwin_set_open_exclusive.restype = None
    _hidapi.hid_darwin_set_open_exclusive(0)


class HIDError(Exception):
    pass


def _enumerate_devices():
    """ Returns all HID devices which are potentially useful to us """
    devices = []
    c_devices = _hidapi.hid_enumerate(0, 0)
    p = c_devices
    while p:
        devices.append(p.contents.as_dict())
        p = p.contents.next
    _hidapi.hid_free_enumeration(c_devices)

    keyboard_or_mouse = {d['path'] for d in devices if d['usage_page'] == 1 and d['usage'] in (6, 2)}
    unique_devices = {}
    for device in devices:
        # On macOS we cannot access keyboard or mouse devices without special permissions. Since
        # we don't need them anyway we remove them so opening them doesn't cause errors later.
        if device['path'] in keyboard_or_mouse:
            # print(f"Ignoring keyboard or mouse device: {device}")
            continue

        # hidapi returns separate entries for each usage page of a device.
        # Deduplicate by path to only keep one device entry.
        if device['path'] not in unique_devices:
            unique_devices[device['path']] = device

    unique_devices = unique_devices.values()
    # print("Unique devices:\n" + '\n'.join([f"{dev}" for dev in unique_devices]))
    return unique_devices


# Use a separate thread to check if devices have been removed or connected
class _DeviceMonitor(Thread):

    def __init__(self, device_callback, polling_delay=5.0):
        self.device_callback = device_callback
        self.polling_delay = polling_delay
        # daemon threads are automatically killed when main thread exits
        super().__init__(daemon=True)

    def run(self):
        # Populate initial set of devices so startup doesn't cause any callbacks
        self.prev_devices = {tuple(dev.items()): dev for dev in _enumerate_devices()}

        # Continously enumerate devices and raise callback for changes
        while True:
            current_devices = {tuple(dev.items()): dev for dev in _enumerate_devices()}
            for key, device in self.prev_devices.items():
                if key not in current_devices:
                    self.device_callback('remove', device)
            for key, device in current_devices.items():
                if key not in self.prev_devices:
                    self.device_callback('add', device)
            self.prev_devices = current_devices
            sleep(self.polling_delay)


# The filterfn is used to determine whether this is a device of interest to Solaar.
# It is given the bus id, vendor id, and product id and returns a dictionary
# with the required hid_driver and usb_interface and whether this is a receiver or device.
def _match(action, device, filterfn):
    vid = device['vendor_id']
    pid = device['product_id']

    # Translate hidapi bus_type to the bus_id values Solaar expects
    if device.get('bus_type') == 0x01:
        bus_id = 0x03  # USB
    elif device.get('bus_type') == 0x02:
        bus_id = 0x05  # Bluetooth
    else:
        bus_id = None

    # Check for hidpp support
    device['hidpp_short'] = False
    device['hidpp_long'] = False
    device_handle = None
    try:
        device_handle = open_path(device['path'])
        report = get_input_report(device_handle, 0x10, 32)
        if len(report) == 1 + 6 and report[0] == 0x10:
            device['hidpp_short'] = True
        report = get_input_report(device_handle, 0x11, 32)
        if len(report) == 1 + 19 and report[0] == 0x11:
            device['hidpp_long'] = True
    except HIDError as e:  # noqa: F841
        if logger.isEnabledFor(logging.INFO):
            logger.info(f"Error opening device {device['path']} ({bus_id}/{vid:04X}/{pid:04X}) for hidpp check: {e}")
    finally:
        if device_handle:
            close(device_handle)

    if logger.isEnabledFor(logging.INFO):
        logger.info(
            'Found device BID %s VID %04X PID %04X HID++ %s %s', bus_id, vid, pid, device['hidpp_short'], device['hidpp_long']
        )

    if not device['hidpp_short'] and not device['hidpp_long']:
        return None

    filter = filterfn(bus_id, vid, pid, device['hidpp_short'], device['hidpp_long'])
    if not filter:
        return
    isDevice = filter.get('isDevice')

    if action == 'add':
        d_info = DeviceInfo(
            path=device['path'].decode(),
            bus_id=bus_id,
            vendor_id=f'{vid:04X}',
            product_id=f'{pid:04X}',
            interface=None,
            driver=None,
            manufacturer=device['manufacturer_string'],
            product=device['product_string'],
            serial=device['serial_number'],
            release=device['release_number'],
            isDevice=isDevice,
            hidpp_short=device['hidpp_short'],
            hidpp_long=device['hidpp_long'],
        )
        return d_info

    elif action == 'remove':
        d_info = DeviceInfo(
            path=device['path'].decode(),
            bus_id=None,
            vendor_id=f'{vid:04X}',
            product_id=f'{pid:04X}',
            interface=None,
            driver=None,
            manufacturer=None,
            product=None,
            serial=None,
            release=None,
            isDevice=isDevice,
            hidpp_short=None,
            hidpp_long=None,
        )
        return d_info


def find_paired_node(receiver_path, index, timeout):
    """Find the node of a device paired with a receiver"""
    return None


def find_paired_node_wpid(receiver_path, index):
    """Find the node of a device paired with a receiver, get wpid from udev"""
    return None


def monitor_glib(callback, filterfn):
    from gi.repository import GLib

    def device_callback(action, device):
        # print(f"device_callback({action}): {device}")
        if action == 'add':
            d_info = _match(action, device, filterfn)
            if d_info:
                GLib.idle_add(callback, action, d_info)
        elif action == 'remove':
            # Removed devices will be detected by Solaar directly
            pass

    monitor = _DeviceMonitor(device_callback=device_callback)
    monitor.start()


def enumerate(filterfn):
    """Enumerate the HID Devices.

    List all the HID devices attached to the system, optionally filtering by
    vendor_id, product_id, and/or interface_number.

    :returns: a list of matching ``DeviceInfo`` tuples.
    """
    for device in _enumerate_devices():
        d_info = _match('add', device, filterfn)
        if d_info:
            yield d_info


def open(vendor_id, product_id, serial=None):
    """Open a HID device by its Vendor ID, Product ID and optional serial number.

    If no serial is provided, the first device with the specified IDs is opened.

    :returns: an opaque device handle, or ``None``.
    """
    if serial is not None:
        serial = ctypes.create_unicode_buffer(serial)

    device_handle = _hidapi.hid_open(vendor_id, product_id, serial)
    if device_handle is None:
        raise HIDError(_hidapi.hid_error(None))
    return device_handle


def open_path(device_path):
    """Open a HID device by its path name.

    :param device_path: the path of a ``DeviceInfo`` tuple returned by enumerate().

    :returns: an opaque device handle, or ``None``.
    """
    if not isinstance(device_path, bytes):
        device_path = device_path.encode()

    device_handle = _hidapi.hid_open_path(device_path)
    if device_handle is None:
        raise HIDError(_hidapi.hid_error(None))
    return device_handle


def close(device_handle):
    """Close a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    """
    assert device_handle
    _hidapi.hid_close(device_handle)


def write(device_handle, data):
    """Write an Output report to a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    :param data: the data bytes to send including the report number as the
    first byte.

    The first byte of data[] must contain the Report ID. For
    devices which only support a single report, this must be set
    to 0x0. The remaining bytes contain the report data. Since
    the Report ID is mandatory, calls to hid_write() will always
    contain one more byte than the report contains. For example,
    if a hid report is 16 bytes long, 17 bytes must be passed to
    hid_write(), the Report ID (or 0x0, for devices with a
    single report), followed by the report data (16 bytes). In
    this example, the length passed in would be 17.

    write() will send the data on the first OUT endpoint, if
    one exists. If it does not, it will send the data through
    the Control Endpoint (Endpoint 0).
    """
    assert device_handle
    assert data
    assert isinstance(data, bytes), (repr(data), type(data))

    bytes_written = _hidapi.hid_write(device_handle, data, len(data))
    if bytes_written < 0:
        raise HIDError(_hidapi.hid_error(device_handle))
    return bytes_written


def read(device_handle, bytes_count, timeout_ms=None):
    """Read an Input report from a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    :param bytes_count: maximum number of bytes to read.
    :param timeout_ms: can be -1 (default) to wait for data indefinitely, 0 to
    read whatever is in the device's input buffer, or a positive integer to
    wait that many milliseconds.

    Input reports are returned to the host through the INTERRUPT IN endpoint.
    The first byte will contain the Report number if the device uses numbered
    reports.

    :returns: the data packet read, an empty bytes string if a timeout was
    reached, or None if there was an error while reading.
    """
    assert device_handle

    data = ctypes.create_string_buffer(bytes_count)
    if timeout_ms is None or timeout_ms < 0:
        bytes_read = _hidapi.hid_read(device_handle, data, bytes_count)
    else:
        bytes_read = _hidapi.hid_read_timeout(device_handle, data, bytes_count, timeout_ms)

    if bytes_read < 0:
        raise HIDError(_hidapi.hid_error(device_handle))
        return None

    return data.raw[:bytes_read]


def get_input_report(device_handle, report_id, size):
    assert device_handle
    data = ctypes.create_string_buffer(size)
    data[0] = bytearray((report_id, ))
    size = _hidapi.hid_get_input_report(device_handle, data, size)
    if size < 0:
        raise HIDError(_hidapi.hid_error(device_handle))
    return data.raw[:size]


def _readstring(device_handle, func, max_length=255):
    assert device_handle
    buf = ctypes.create_unicode_buffer(max_length)
    ret = func(device_handle, buf, max_length)
    if ret < 0:
        raise HIDError('Error reading device property')
    return buf.value


def get_manufacturer(device_handle):
    """Get the Manufacturer String from a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    """
    return _readstring(device_handle, _hidapi.get_manufacturer_string)


def get_product(device_handle):
    """Get the Product String from a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    """
    return _readstring(device_handle, _hidapi.get_product_string)


def get_serial(device_handle):
    """Get the serial number from a HID device.

    :param device_handle: a device handle returned by open() or open_path().
    """
    return _readstring(device_handle, _hidapi.get_serial_number_string)



# -*- python-mode -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging

import yaml as _yaml

import gi  # isort:skip

gi.require_version('Gtk', '3.0')  # NOQA: E402
from gi.repository import GLib, Gtk, Gio  # NOQA: E402 # isort:skip
from logitech_receiver.status import ALERT  # NOQA: E402 # isort:skip
from solaar.i18n import _  # NOQA: E402 # isort:skip

logger = logging.getLogger(__name__)

#
#
#

assert Gtk.get_major_version() > 2, 'Solaar requires Gtk 3 python bindings'

GLib.threads_init()

#
#
#


def _error_dialog(reason, object):
    logger.error('error: %s %s', reason, object)

    if reason == 'permissions':
        title = _('Permissions error')
        text = (
            _('Found a Logitech receiver or device (%s), but did not have permission to open it.') % object + '\n\n' +
            _("If you've just installed Solaar, try disconnecting the receiver or device and then reconnecting it.")
        )
    elif reason == 'nodevice':
        title = _('Cannot connect to device error')
        text = (
            _('Found a Logitech receiver or device at %s, but encountered an error connecting to it.') % object + '\n\n' +
            _('Try disconnecting the device and then reconnecting it or turning it off and then on.')
        )
    elif reason == 'unpair':
        title = _('Unpairing failed')
        text = (
            _('Failed to unpair %{device} from %{receiver}.').format(device=object.name, receiver=object.receiver.name) +
            '\n\n' + _('The receiver returned an error, with no further details.')
        )
    else:
        raise Exception("ui.error_dialog: don't know how to handle (%s, %s)", reason, object)

    assert title
    assert text

    m = Gtk.MessageDialog(None, Gtk.DialogFlags.MODAL, Gtk.MessageType.ERROR, Gtk.ButtonsType.CLOSE, text)
    m.set_title(title)
    m.run()
    m.destroy()


def error_dialog(reason, object):
    assert reason is not None
    GLib.idle_add(_error_dialog, reason, object)


#
#
#

_task_runner = None


def ui_async(function, *args, **kwargs):
    if _task_runner:
        _task_runner(function, *args, **kwargs)


#
#
#

from . import diversion_rules, notify, tray, window  # isort:skip  # noqa: E402


def _startup(app, startup_hook, use_tray, show_window):
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('startup registered=%s, remote=%s', app.get_is_registered(), app.get_is_remote())

    from solaar.tasks import TaskRunner as _TaskRunner
    global _task_runner
    _task_runner = _TaskRunner('AsyncUI')
    _task_runner.start()

    notify.init()
    if use_tray:
        tray.init(lambda _ignore: window.destroy())
    window.init(show_window, use_tray)

    startup_hook()


def _activate(app):
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('activate')
    if app.get_windows():
        window.popup()
    else:
        app.add_window(window._window)


def _command_line(app, command_line):
    args = command_line.get_arguments()
    args = _yaml.safe_load(''.join(args)) if args else args
    if not args:
        _activate(app)
    elif args[0] == 'config':  # config call from remote instance
        if logger.isEnabledFor(logging.INFO):
            logger.info('remote command line %s', args)
        from solaar.ui.config_panel import change_setting  # prevent circular import
        from solaar.ui.window import find_device  # prevent circular import
        dev = find_device(args[1])
        if dev:
            setting = next((s for s in dev.settings if s.name == args[2]), None)
            if setting:
                change_setting(dev, setting, args[3:])
    return 0


def _shutdown(app, shutdown_hook):
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('shutdown')

    shutdown_hook()

    # stop the async UI processor
    global _task_runner
    _task_runner.stop()
    _task_runner = None

    tray.destroy()
    notify.uninit()


def run_loop(startup_hook, shutdown_hook, use_tray, show_window):
    assert use_tray or show_window, 'need either tray or visible window'
    # from gi.repository.Gio import ApplicationFlags as _ApplicationFlags
    APP_ID = 'io.github.pwr_solaar.solaar'
    application = Gtk.Application.new(APP_ID, Gio.ApplicationFlags.HANDLES_COMMAND_LINE)

    application.connect('startup', lambda app, startup_hook: _startup(app, startup_hook, use_tray, show_window), startup_hook)
    application.connect('command-line', _command_line)
    application.connect('activate', _activate)
    application.connect('shutdown', _shutdown, shutdown_hook)

    application.register()
    if application.get_is_remote():
        print(_('Another Solaar process is already running so just expose its window'))
    application.run()


#
#
#


def _status_changed(device, alert, reason, refresh=False):
    assert device is not None
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('status changed: %s (%s) %s', device, alert, reason)

    tray.update(device)
    if alert & ALERT.ATTENTION:
        tray.attention(reason)

    need_popup = alert & ALERT.SHOW_WINDOW
    window.update(device, need_popup, refresh)
    diversion_rules.update_devices()

    if alert & (ALERT.NOTIFICATION | ALERT.ATTENTION):
        notify.show(device, reason)


def status_changed(device, alert=ALERT.NONE, reason=None, refresh=False):
    GLib.idle_add(_status_changed, device, alert, reason, refresh)
